const fs = require('fs');
const path = require('path');
const versionUtils = require('../utils/VersionUtils');
const ioUtils = require('../utils/IOUtils');
const md5File = require('md5-file');

class NodePluginExporter {

    constructor(dist) {
        this.dist = dist;
        this.bins = {};
        this.plugins = {};
        this.optionalPlugins = {};
    }

    getPluginPath(pluginName) {
        let pluginPath = null;
        let dirs = module.paths;
        for (let dir of dirs) {
            let pluginJsonPath = path.resolve(dir, pluginName, "package.json");
            if (fs.existsSync(pluginJsonPath)) {
                pluginPath = path.resolve(dir, pluginName);
                break;
            }
        }
        return pluginPath;
    }

    isNeedToUpdate(pluginName, comparePluginPath) {
        let distDir = path.resolve(this.dist, "node_modules", pluginName);
        let distPackJsonPath = path.resolve(distDir, "package.json");
        let needUpdate = true;
        if (fs.existsSync(distPackJsonPath)) {
            let packageJson = require(distPackJsonPath);
            let selfPackjson = require(path.resolve(comparePluginPath, "package.json"));
            needUpdate = versionUtils.isNew(selfPackjson.version, packageJson.version);
        }
        return needUpdate;
    }

    iter(pluginName, pluginPath) {
        pluginPath = pluginPath || this.getPluginPath(pluginName);
        if (pluginPath != null) {
            let packageJson = require(path.resolve(pluginPath, "package.json"));
            let bin = packageJson.bin;
            if (bin) {
                Object.keys(bin).forEach(binName => {
                    this.bins[binName] = true;
                });
            }
            let dependencies = packageJson.dependencies;
            if (dependencies) {
                let optionalDependencies = packageJson.optionalDependencies;
                Object.keys(dependencies).forEach(depName => {
                    if (!this.plugins[depName]) {
                        let childPath = this.getPluginPath(depName);
                        if (optionalDependencies && optionalDependencies[depName]) {
                            this.optionalPlugins[depName] = childPath;
                        } else {
                            this.plugins[depName] = childPath;
                        }
                        this.iter(depName, childPath);
                    }
                });
            }
            let selfNodeModulesPath = path.resolve(pluginPath, "node_modules");
            if (fs.existsSync(selfNodeModulesPath)) {
                let fileNames = fs.readdirSync(selfNodeModulesPath);
                if (fileNames && fileNames.length > 0) {
                    fileNames.forEach(fileName => {
                        let childPath = path.resolve(selfNodeModulesPath, fileName);
                        let childPackageJson = path.resolve(childPath, "package.json");
                        if (fs.existsSync(childPackageJson)) {
                            this.iter(fileName, childPath);
                        }
                    });
                }
            }
        }
    }

    export (pluginName) {
        return new Promise((resolve, reject) => {
            let pluginPath = this.getPluginPath(pluginName);
            this.plugins[pluginName] = pluginPath;
            this.iter(pluginName, pluginPath);
            let promises = [];
            let nodeModules = path.resolve(fs.realpathSync('.'), "node_modules");
            let binSourcePath = path.resolve(nodeModules, '.bin');
            let modulesDistPath = path.resolve(this.dist, "node_modules");
            let binDistPath = path.resolve(modulesDistPath, ".bin");
            let promise = ioUtils.copyDirWithFilter(binSourcePath, binDistPath, (srcPath, distPath) => {
                let fileName = path.basename(srcPath);
                let index = fileName.lastIndexOf('.');
                let fileNameWithoutExt = index == -1 ? fileName : fileName.substring(0, index);
                if (this.bins[fileNameWithoutExt]) {
                    if (!fs.existsSync(distPath)) {
                        return true;
                    } else {
                        return md5File.sync(srcPath) !== md5File.sync(distPath);
                    }
                }
                return false;
            });
            promises.push(promise);

            for (let [pluginName, pluginPath] of Object.entries(this.plugins)) {
                if (pluginPath != null) {
                    let needUpdate = this.isNeedToUpdate(pluginName, pluginPath);
                    if (needUpdate) {
                        let promise = ioUtils.copyDirWithFilter(pluginPath, path.resolve(modulesDistPath, pluginName));
                        promises.push(promise);
                    }
                }
            }

            for (let [pluginName, pluginPath] of Object.entries(this.optionalPlugins)) {
                if (pluginPath != null) {
                    let needUpdate = this.isNeedToUpdate(pluginName, pluginPath);
                    if (needUpdate) {
                        let promise = ioUtils.copyDirWithFilter(pluginPath, path.resolve(modulesDistPath, pluginName));
                        promises.push(promise);
                    }
                }
            }

            Promise.all(promises).then(() => {
                resolve();
            }).catch(err => {
                reject(err);
            });
        });
    }

}

module.exports = NodePluginExporter;